/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#include "imageinfo.h"
#include <iostream>

ImageInfo::Handle ImageInfo::clone()
{
  ImageInfo::Handle clone = ImageInfo::Handle(new ImageInfo(m_x,m_y,m_z,m_v,DT_SIGNED_SHORT,
							    m_xDim,m_yDim,m_zDim,
							    m_tr,m_bitsPerVoxel,m_imageName + "-mask",m_dtiImage));

  clone->setXOrig(inqXOrig());
  clone->setYOrig(inqYOrig());
  clone->setZOrig(inqZOrig());

  clone->setReadOnly(false);
  clone->setTarnished(false);

  return ImageInfo::Handle(clone);
}

ImageInfo::ImageInfo(short x,short y,short z,short v,short dt,
                     float xDim, float yDim, float zDim,float tr,
                     size_t bitsPerVoxel,std::string name,
                     bool dtiImage):
  m_x(x),m_y(y),m_z(z),m_v(v),m_dt(dt),
  m_xDim(xDim),m_yDim(yDim),m_zDim(zDim),
  m_tr(tr),m_imageName(name),m_bitsPerVoxel(bitsPerVoxel),
  m_dtiImage(dtiImage), m_mainImage(false), m_isStoredRadiological(true)
{
  int i,j;
  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
      m_stdmat.m[i][j]=0;
    }
  }
  m_stdmat.m[0][0]=fabs(xDim);
  m_stdmat.m[1][1]=fabs(yDim);
  m_stdmat.m[2][2]=fabs(zDim);
  m_stdmat.m[3][3]=1;
}


ImageInfo::ImageInfo(FSLIO *avw,std::string filename):m_fileName(filename)
{
  m_bitsPerVoxel = FslGetDataType(avw, &m_dt);
  FslGetDim(avw, &m_x, &m_y, &m_z, &m_v);
  FslGetVoxDim(avw, &m_xDim, &m_yDim, &m_zDim, &m_tr);
  FslGetOriginator(avw, &m_orig[0]);
  FslGetAuxFile(avw,&m_auxFile[0]);
  if(strlen(m_auxFile) != 0) {
    FslGetCalMinMax(avw, &m_min, &m_max);
  } else {
    int glMin, glMax;
    FslGetMinMax(avw, &glMin, &glMax);
    m_min = glMin;
    m_max = glMax;
  }

  FslGetStdXform(avw,&m_stdmat);
  
  m_isStoredRadiological = FslGetLeftRightOrder(avw) == FSL_RADIOLOGICAL;

  m_imageName = extractName(m_fileName);
  m_readOnly  = true;
  m_tarnished = false;
  m_dtiImage = false;
  m_mainImage = false;
}

ImageInfo::~ImageInfo()
{
}

/*
 * Commit the Header information and write it back to disc
 */
void ImageInfo::saveAvwHeader(FSLIO *avw)
{
  FslSetDataType(avw, m_dt);
  FslSetDim(avw, m_x, m_y, m_z, m_v);
  FslSetVoxDim(avw, m_xDim, m_yDim, m_zDim, m_tr);
  FslSetAuxFile(avw, &m_auxFile[0]);
  FslSetOriginator(avw, &m_orig[0]);
  FslSetCalMinMax(avw, m_min, m_max);
  FslSetMinMax(avw, int(m_min), int(m_max));
  FslWriteHeader(avw);
}

void ImageInfo::setMin(float min) { m_min = min; }
void ImageInfo::setMax(float max) { m_max = max; }
float ImageInfo::inqMin() const { return m_min; }
float ImageInfo::inqMax() const { return m_max; }

std::string ImageInfo::inqLutName() const 
{ 
  return std::string(m_auxFile);
}

void ImageInfo::setLutName(std::string lutName)
{
  lutName.copy(m_auxFile,24);
}

void ImageInfo::setImageName(std::string imageName)
{
  m_imageName = imageName;
}

bool ImageInfo::isCompatible(ImageInfo::Handle other) const
{
  return ((other->inqX() == m_x) &&
          (other->inqY() == m_y) &&
          (other->inqZ() == m_z));
}

bool ImageInfo::isValidCoordinate(short x, short y, short z) const
{
  return(((x >= 0)&&(x < m_x)) &&
         ((y >= 0)&&(y < m_y)) &&
         ((z >= 0)&&(z < m_z)));
}

std::string ImageInfo::extractName(std::string filename)
{
  std::string result;
  std::string ext;

  result.erase();
  result = filename.substr(filename.rfind('/')+ 1,filename.length()- 1);

  return result;
} 

bool ImageInfo::isDtiCompatible() const
{
  return (m_v > 2);
}

bool ImageInfo::isInteger() const
{
  return (m_dt < 16);
}

bool  ImageInfo::inqNoDimensions() const
{
  return (m_xDim == 0.0 || m_yDim == 0.0 || m_zDim == 0.0);
}

void ImageInfo::setStdMat(mat44 stdmat) { m_stdmat = stdmat; }

/*
bool ImageInfo::isStandard() const
{
  return ( (m_x == 91) && (m_y == 109) && (m_z == 91) &&
	   (m_xDim == -2) && (m_yDim == 2) && (m_zDim == 2) &&
	   (inqXOrig() == 46) && (inqYOrig() == 64) && (inqZOrig() == 37) );
}
*/

