/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#if !defined(IMAGEINFO_H)
#define IMAGEINFO_H

#include <boost/shared_ptr.hpp>
#include <string>

#include "fslio/fslio.h"

class ImageInfo
{
public:
  typedef boost::shared_ptr< ImageInfo > Handle;

  ~ImageInfo();

  short inqX() const;
  short inqY() const;
  short inqZ() const;
  short inqNumVolumes() const;
  bool  inqNoDimensions() const;
  float inqXDim() const;
  float inqYDim() const;
  float inqZDim() const;
  short inqXOrig() const;
  short inqYOrig() const;
  short inqZOrig() const;
  void setXOrig(short);
  void setYOrig(short);
  void setZOrig(short);

  float inqTr() const;
  short inqDt() const;
  size_t inqBytesPerVoxel() const;
  size_t inqBitsPerVoxel() const;
  std::string inqImageName() const;
  std::string inqFileName() const;
  std::string inqLutName() const;
  bool inqReadOnly() const;
  bool isStoredRadiological() const;
  bool isCompatible(ImageInfo::Handle) const;
  bool isValidCoordinate(short x, short y, short z) const;
  bool isMainImage() const;
  bool isInteger() const;
  bool isDtiCompatible() const;
  bool isDtiImage() const;
  void setDtiImage(bool);
  void setImageName(std::string);
  void setReadOnly(bool);
  void setAsMainImage();
  void setLutName(std::string);
  void setTarnished(bool state);
  bool inqTarnished() const;     

  /** Image min instensity
   *  @return The stored minimum value or the minimum from the header if not yet set.
   */
  float inqMin() const;
  float inqMax() const;
  void setMin(float);
  void setMax(float);

  mat44 inqStdMat() const;
  void setStdMat(mat44 stdmat);

  static ImageInfo::Handle init(FSLIO *avw, std::string filename) {
    return ImageInfo::Handle( new ImageInfo(avw,filename) );
  }

  void saveAvwHeader(FSLIO *avw);

  Handle clone();

private:
  ImageInfo(FSLIO *avw,std::string filename);
  ImageInfo(short x,short y,short z,short v,short dt,
            float xDim, float yDim, float zDim,float tr,
            size_t bitsPerVoxel,std::string name, bool dtiImage);
  std::string extractName(std::string filename);


  bool m_minSet;
  bool m_maxSet;
  float m_min;
  float m_max;

  short m_x;
  short m_y;
  short m_z;
  short m_v;
  short m_dt;
  float m_xDim;
  float m_yDim;
  float m_zDim;
  float m_tr;
  short m_orig[5];
  char  m_auxFile[24];
  std::string m_fileName;
  std::string m_imageName;
 
  mat44 m_stdmat;

  size_t m_bitsPerVoxel;
  
  short m_type;  
  bool  m_readOnly;
  bool  m_tarnished;
  bool  m_dtiImage;
  bool  m_mainImage;
  bool  m_isStoredRadiological;
};

inline void  ImageInfo::setTarnished(bool state) {m_tarnished = state;}
inline bool  ImageInfo::inqTarnished() const     {return m_tarnished;}

inline void  ImageInfo::setReadOnly(bool state) {m_readOnly = state;}
inline bool  ImageInfo::inqReadOnly() const {return m_readOnly;}

inline void  ImageInfo::setDtiImage(bool state) {m_dtiImage = state;}
inline bool  ImageInfo::isDtiImage() const {return m_dtiImage;}

inline short ImageInfo::inqX() const { return m_x; }
inline short ImageInfo::inqY() const { return m_y; }
inline short ImageInfo::inqZ() const { return m_z; }
inline short ImageInfo::inqNumVolumes() const { return m_v; }

inline float ImageInfo::inqXDim() const { return m_xDim; }
inline float ImageInfo::inqYDim() const { return m_yDim; }
inline float ImageInfo::inqZDim() const { return m_zDim; }
inline short ImageInfo::inqXOrig() const { return m_orig[0];}
inline short ImageInfo::inqYOrig() const { return m_orig[1];}
inline short ImageInfo::inqZOrig() const { return m_orig[2];}
inline void ImageInfo::setXOrig(short v) { m_orig[0] = v;}
inline void ImageInfo::setYOrig(short v) { m_orig[1] = v;}
inline void ImageInfo::setZOrig(short v) { m_orig[2] = v;}
inline float ImageInfo::inqTr() const { return m_tr; }
inline short ImageInfo::inqDt() const { return m_dt; }

inline void ImageInfo::setAsMainImage(){m_mainImage = true;}
inline bool ImageInfo::isMainImage()const {return m_mainImage;}

inline std::string ImageInfo::inqImageName() const { return m_imageName; }
inline std::string ImageInfo::inqFileName() const { return m_fileName; }
inline size_t ImageInfo::inqBytesPerVoxel() const { return m_bitsPerVoxel / 8; }
inline size_t ImageInfo::inqBitsPerVoxel() const { return m_bitsPerVoxel; }

inline mat44 ImageInfo::inqStdMat() const { return m_stdmat; }

inline bool ImageInfo::isStoredRadiological() const { return m_isStoredRadiological; }

#endif
