/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#if !defined(CURSOR_H)
#define CURSOR_H

#include <list>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>

class CursorObserver;

/** Cursor
 * An observable cursor class @see Design Patterns p293
 */
class Cursor
{
public:
  typedef boost::shared_ptr< Cursor > Handle;
  typedef boost::weak_ptr< Cursor > WeakHandle;
  
  Handle clone();

  void setCursor(short x, short y, short z, short v);
  void setCursor(short x, short y, short z);
  void setCursorRepaint(short x, short y, short z);
  void repaint();
  bool  inqRepaint() const;
  short inqX() const;
  short inqY() const;
  short inqZ() const;
  short inqV() const;

  /** Create a new cursor
   * @param o the initial image which will follow this cursor
   */
  static Handle create(CursorObserver *o, 
                       short xMax, short yMax, short zMax,short vMax);

  /** Attach a view to this cursor
   * @param o handle of a view which requires notification of any 
   *          changes to this cursor
   */
  void attach(CursorObserver *o);

  /** Detach a view from this cursor
   * @param o handle of the view to be removed from the notification list
   */
  void detach(CursorObserver *o);

  virtual ~Cursor() {}

protected:
  Handle countedThis() const { return boost::make_shared(m_countedThis); }

private:
  Cursor(CursorObserver *o,short xMax, short yMax, short zMax, short vMax);
  Cursor(short x,    short y,    short z,    short v,
         short xMax, short yMax, short zMax, short vMax);

  void setCountedThis(const Handle c) { m_countedThis = WeakHandle(c); }

  /** Notify all observers of a change to this cursor
   */
  void notify() const;
  void setXYZ(short,short,short);
  void setV(short);
  short m_x;
  short m_y;
  short m_z;
  short m_v;
  short m_xMax;
  short m_yMax;
  short m_zMax;
  short m_vMax;

  bool  m_repaint;

  WeakHandle m_countedThis;

  std::list< CursorObserver * > m_observers;
};

inline short Cursor::inqX() const { return m_x; }
inline short Cursor::inqY() const { return m_y; }
inline short Cursor::inqZ() const { return m_z; }
inline short Cursor::inqV() const { return m_v; }
inline bool  Cursor::inqRepaint() const {return m_repaint;}

class CursorObserver
{
public:
  virtual ~CursorObserver() {}

  virtual void update(const Cursor::Handle& c) = 0;

  CursorObserver() {}
};


#endif
