/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */

#include <qlineedit.h>
#include <qspinbox.h>
#include <qtoolbutton.h>
#include <qpixmap.h>
#include <qvalidator.h>
#include <qtooltip.h>
#include <qpixmap.h>
#include <qlayout.h>
#include <algorithm>
#include <stdio.h>

#include "fill.xpm"
#include "undo.xpm"
#include "pen.xpm"
#include "eraser.xpm"

#include "tracker.h"
#include "drawwidget.h"

const char * sizeBoxText = "Pen size.<br><hr>Changes the dimensions of the masking pen.";
const char *  valBoxText = "Pen value<br><hr>Controls the value that is written when drawing a mask.";
const char *  undoButtonText = "Undo button<br><hr>Reverts mask back to state before last edit.";
const char *  penButtonText = "Pen button<br><hr>Selects pen mode for drawing pixels on images.";
const char *  eraseButtonText = "Eraser button<br><hr>Selects erase mode for removing pixels from images. Pressing the Control key in Masking Mode also toggles the Eraser button";
const char *  fillButtonText = "Fill button<br><hr>Selects fill mode for filling enclosed shapes.";

DrawWidget::DrawWidget(QWidget* w) :   QWidget(w), m_mode(Pen), 
                                       m_penSize(1),  m_penValue(1),
                                       m_eraseSize(1),m_eraseValue(0),
                                       m_fillSize(1), m_fillValue(1),
                                       m_undoPressed(false), m_prevMode(Pen)
{

  m_valBox = new QSpinBox(this);
  m_valBox->setValidator(new QIntValidator(this));
  m_valBox->setRange(0,255);
  
  connect(m_valBox,SIGNAL(valueChanged(int)),
          this,SLOT(valBoxChanged(int)));
  
  m_sizeBox = new QSpinBox(this);
  m_sizeBox->setValidator(new QIntValidator(this));
  m_sizeBox->setRange(1,20);


  connect(m_sizeBox,SIGNAL(valueChanged(int)),
          this,SLOT(sizeBoxChanged(int)));
 
  m_undoButton = new QToolButton(this);
  m_undoButton->setPixmap(QPixmap(undo));
  m_undoButton->setMaximumSize(QPixmap(undo).size());
  m_undoButton->setAccel(CTRL+Key_Z);

  connect(m_undoButton, SIGNAL(pressed()),this, SLOT(undoButtonPressed())); 
 
  m_fillButton = new QToolButton(this);
  m_fillButton->setPixmap(QPixmap(fill_xpm));
  m_fillButton->setMaximumSize(QPixmap(fill_xpm).size());
  m_fillButton->setToggleButton(true);

  connect(m_fillButton, SIGNAL(toggled(bool)),this, SLOT(fillButtonToggled(bool)));

  m_penButton = new QToolButton(this);
  m_penButton->setPixmap(QPixmap(pen_xpm));
  m_penButton->setMaximumSize(QPixmap(pen_xpm).size());
  m_penButton->setToggleButton(true);

  connect(m_penButton, SIGNAL(toggled(bool)),this, SLOT(penButtonToggled(bool))); 

  m_eraseButton = new QToolButton(this);
  m_eraseButton->setPixmap(QPixmap(eraser));
  m_eraseButton->setMaximumSize(QPixmap(eraser).size());
  m_eraseButton->setToggleButton(true);

  connect(m_eraseButton, SIGNAL(toggled(bool)),this, SLOT(eraseButtonToggled(bool))); 


  QToolTip::add(m_valBox,     tr(valBoxText));
  QToolTip::add(m_sizeBox,    tr(sizeBoxText));
  QToolTip::add(m_undoButton, tr(undoButtonText));
  QToolTip::add(m_penButton,  tr(penButtonText));
  QToolTip::add(m_eraseButton,tr(eraseButtonText));
  QToolTip::add(m_fillButton, tr(fillButtonText));
 
  QHBoxLayout* hlayout = new QHBoxLayout(this);

  hlayout->addWidget(m_valBox);
  hlayout->addWidget(m_sizeBox);
  hlayout->addWidget(m_penButton);
  hlayout->addWidget(m_eraseButton); 
  hlayout->addWidget(m_fillButton);
  hlayout->addWidget(m_undoButton);


  show();

  updateControls();
}

DrawWidget::~DrawWidget()
{

}


void DrawWidget::attach(DrawWidgetObserver* o)
{
  m_observers.remove(o);
  m_observers.push_back(o);
}

void DrawWidget::detach(DrawWidgetObserver* o)
{
  m_observers.remove(o);
}

struct Update
{
  Update(DrawWidget* l): m_ol(l) {}

  void operator()(DrawWidgetObserver* v)
  {
    v->update(m_ol);
  }

  DrawWidget* m_ol;
};

void DrawWidget::notify()
{
    std::for_each(m_observers.begin(), m_observers.end(), Update(this));
}

void DrawWidget::sizeBoxChanged(int value)
{
  if(inqMode() == Pen)   m_penSize   = m_sizeBox->value();
  if(inqMode() == Fill)  m_penSize   = m_sizeBox->value();
  if(inqMode() == Erase) m_eraseSize = m_sizeBox->value();
}

void DrawWidget::valBoxChanged(int value)
{  
  if(inqMode() == Pen)   m_penValue   = m_valBox->value();
  if(inqMode() == Fill)   m_penValue   = m_valBox->value();
  if(inqMode() == Erase) m_eraseValue = m_valBox->value();
}

void DrawWidget::undoButtonPressed()
{
  m_undoPressed = true;
  notify();
  m_undoPressed = false;
}

void DrawWidget::fillButtonToggled(bool state)
{
  //  if(state){ m_mode = Fill; notify();}
  //  updateControls();

  if(state){setMode(Fill);}
}

void DrawWidget::penButtonToggled(bool state)
{
  //  if(state){ m_mode = Pen; notify();}
  //  updateControls();  

  if(state){setMode(Pen);}

}

void DrawWidget::eraseButtonToggled(bool state)
{
  //  if(state){ m_mode = Erase; notify();}
  //  updateControls();  

  if(state){setMode(Erase);}

}

void DrawWidget::updateControls()
{  

  blockControlSignals(true);
  switch(inqMode())
    {
    case Pen:
      m_penButton->setOn(true);
      m_eraseButton->setOn(false);     
      m_fillButton->setOn(false);       
      m_valBox->setValue(m_penValue);
      m_sizeBox->setValue(m_penSize); 
      break;
    case Erase:    
      m_eraseButton->setOn(true);      
      m_penButton->setOn(false);      
      m_fillButton->setOn(false);         
      m_valBox->setValue(m_eraseValue);
      m_sizeBox->setValue(m_eraseSize);      
      break;
    case Fill:
      m_fillButton->setOn(true);
      m_penButton->setOn(false);
      m_eraseButton->setOn(false);      
      if(m_prevMode == Pen)
        {
          m_fillValue = m_penValue;
          m_fillSize  = m_penSize; 
        }
      else 
        if(m_prevMode == Erase)
        {
          m_fillValue = m_eraseValue;
          m_fillSize  = m_eraseSize; 
        }
      m_valBox->setValue(m_fillValue);
      m_sizeBox->setValue(m_fillSize);

      break; 
    }
  blockControlSignals(false);
}

int DrawWidget::inqPixelSize() 
{  
  int size(0);

  switch(inqMode())
    {
    case Pen:
      size = m_penSize;
      break;
    case Erase:
      size = m_eraseSize;
      break;
    case Fill:
      size = m_fillSize;
      break;
    }
  return size;
}

int DrawWidget::inqPixelValue()
{
  int value(0);

  switch(inqMode())
    {
    case Pen:
      value = m_penValue;
      break;
    case Erase:
      value = m_eraseValue;
      break;
    case Fill:
      value = m_fillValue;
      break;
    }
  return value;
}

void DrawWidget::setMode(int m)
{
  m_prevMode = m_mode;
  m_mode = m;
  notify();
  updateControls();
}

void DrawWidget::setPrevMode()
{
  setMode(m_prevMode);
}

void DrawWidget::blockControlSignals(bool state)
{
  m_valBox->blockSignals(state); 
  m_sizeBox->blockSignals(state);
  m_undoButton->blockSignals(state);
  m_penButton->blockSignals(state);
  m_eraseButton->blockSignals(state);
  m_fillButton->blockSignals(state);
}
