/*  FSLView - 2D/3D Interactive Image Viewer

    James Saunders, David Flitney and Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 2002-2003 University of Oxford  */

/*  CCOPYRIGHT */


#include "imagegroup.h"
#include "imagedisplaysetting.h"
#include "tracker.h"
#include <qobject.h>

#include <list>

class CheckImageTarnish{
public:
  CheckImageTarnish():m_tarnished(false){}
  void operator()(Image::Handle image) 
  {
    m_tarnished = image->getInfo()->inqTarnished();
  }
  bool result(){return m_tarnished;}
private:
  bool m_tarnished;
};



class LutSearch {
public:
  LutSearch(std::string const & name) : m_name(name) {}
  void operator()(LookUpTable::Handle lut) 
  {
    if(lut->inqLutName() ==  m_name){m_lut = lut;}
  }

  LookUpTable::Handle result(){return m_lut;}
private:
  std::string m_name;  
  LookUpTable::Handle m_lut;
};

struct ImageGroup::Implementation
{
  Implementation(Image::Handle image){
     m_xDim = m_yDim = m_zDim = 0;
     m_initLutSize = 0;
     setImage(image);
  }
  void setImage(Image::Handle image) {
    image->getInfo()->setAsMainImage();
    m_imageList.push_back(image);
   
    m_xDim = m_imageList.front()->getVolume(0)->inqX();
    m_yDim = m_imageList.front()->getVolume(0)->inqY();
    m_zDim = m_imageList.front()->getVolume(0)->inqZ();
    m_min  = m_imageList.front()->getInfo()->inqMin();
    m_max  = m_imageList.front()->getInfo()->inqMax();
  }

 
  ~Implementation(){TRACKER("ImageGroup::~Implementation()");}
  
  std::list<ImageGroupObserver* >  m_observers;
  std::vector<Image::Handle>         m_imageList;
  std::vector<LookUpTable::Handle>   m_lookUpTables;
  int  m_initLutSize;
  int  m_xDim;
  int  m_yDim;
  int  m_zDim;
  float  m_min;
  float  m_max;
};

ImageGroup::ImageGroup(Image::Handle image):m_impl(new Implementation(image))
{
  TRACKER("ImageGroup::ImageGroup(Image::Handle& image)");
  addLookUpTable(LookUpTable::redYellow());
  addLookUpTable(LookUpTable::blueLightblue());
  addLookUpTable(LookUpTable::red());
  addLookUpTable(LookUpTable::blue());
  addLookUpTable(LookUpTable::green());
  addLookUpTable(LookUpTable::yellow());
  addLookUpTable(LookUpTable::render1());
  addLookUpTable(LookUpTable::render1t());
  addLookUpTable(LookUpTable::render2());
  addLookUpTable(LookUpTable::render2t());
  addLookUpTable(LookUpTable::render3());

  m_impl->m_initLutSize = m_impl->m_lookUpTables.size();
}

ImageGroup::~ImageGroup()
{
  TRACKER("ImageGroup::~ImageGroup()");
}

/** 
 * The create methods are in-lieu of conventional constructors.
 * Generate a new image group with a given image. 
 * 
 * @param image Main image for new image group
 * 
 * @return Reference counted handle to a new ImageGroup.
 */
ImageGroup::Handle ImageGroup::create(Image::Handle image)
{
  return ImageGroup::Handle(new ImageGroup(image));
}

/** 
 * Add an overlay to this image group
 * 
 * @param image New image to be added to the image groups overlay list.
 * 
 * @return true if operation succeeds.
 */
bool ImageGroup::addOverlay(Image::Handle image)
{
  m_impl->m_imageList.push_back(image);
  notify(ImageGroupMsg(NewOverlay));
  return true;
}

/** 
 * Add a lut to the image groups look up table list.
 * 
 * @param lut New lut to be added to the lut list.
 */
void ImageGroup::addLookUpTable(LookUpTable::Handle lut)
{
  m_impl->m_lookUpTables.push_back(lut);
  notify(ImageGroupMsg(NewLookUpTable));
}

Image::Handle ImageGroup::getLatestImage() const
{
  return m_impl->m_imageList.back();
}

LookUpTable::Handle ImageGroup::getLatestLUT()
{
  LookUpTable::Handle latest;
  
  if(!m_impl->m_lookUpTables.empty())
  {
    latest = m_impl->m_lookUpTables.back();
  }

  return latest;
}

Image::Handle ImageGroup::getImage(int n) const
{
  Image::Handle img;
  int size = m_impl->m_imageList.size();

  if((size > 0) && (n < size))
    img = m_impl->m_imageList[n];

  return img;
}

LookUpTable::Handle ImageGroup::getLut(int n) const
{
  LookUpTable::Handle lut;
  int size = m_impl->m_lookUpTables.size();

  if((size > 0) && (n < size))
    lut = m_impl->m_lookUpTables[n];

  return lut;
}

LookUpTable::Handle ImageGroup::getLut(std::string const &name) const
{
  LutSearch search = std::for_each(m_impl->m_lookUpTables.begin(),
                                   m_impl->m_lookUpTables.end(),
                                   LutSearch(name));

  return search.result();
}

int ImageGroup::getInitialLutCount() const
{
  return m_impl->m_initLutSize;
}

/** 
 * Remove a given overlay from the image group. Currently will only
 * remove the last overlay if there is one.
 * 
 * @param image The image handle for the overlay image to be removed.
 * 
 * @return true if operation succeeded.
 */
bool ImageGroup::remOverlay(Image::Handle image)
{
  bool result(false);

  if (!m_impl->m_imageList.empty())
    {     

      ImageList::iterator cur = std::find(m_impl->m_imageList.begin(),
                                          m_impl->m_imageList.end(),
                                          image);
      if(cur != m_impl->m_imageList.end())
        {
          m_impl->m_imageList.erase(cur);
          notify(ImageGroupMsg(RemOverlay));
          result = true;
        }
    }

  return result;
}

/** 
 * Access the groups main image.
 * 
 * @return Image::Handle for the groups main image.
 */
Image::Handle ImageGroup::getMainImage()
{
  return m_impl->m_imageList.front();
}

ImageGroup::ImageList::iterator  ImageGroup::begin()
{
  return m_impl->m_imageList.begin();
}

ImageGroup::ImageList::iterator  ImageGroup::end()
{
  return m_impl->m_imageList.end();
}

ImageGroup::LutList::iterator ImageGroup::beginLutList()
{
  return m_impl->m_lookUpTables.begin();
}

ImageGroup::LutList::iterator ImageGroup::endLutList()
{
  return m_impl->m_lookUpTables.end();
}

int  ImageGroup::size()
{
  return m_impl->m_imageList.size();
}

int ImageGroup::inqX()
{
  return m_impl->m_xDim;
}

int ImageGroup::inqY()
{ 
   return m_impl->m_yDim; 
}

int ImageGroup::inqZ()
{   
  return m_impl->m_zDim;   
}

float ImageGroup::inqMin() { return m_impl->m_min; }

float ImageGroup::inqMax() { return m_impl->m_max; }

void ImageGroup::attach(ImageGroupObserver* o)
{
   m_impl->m_observers.remove(o);
   m_impl->m_observers.push_back(o);
}

void ImageGroup::detach(ImageGroupObserver* o)
{
   m_impl->m_observers.remove(o);
}

struct Update
{
  Update(ImageGroup* i, ImageGroupMsg m): m_imageGroup(i),m_msg(m) {}

  void operator()(ImageGroupObserver* v)
  {
    v->update(m_imageGroup,m_msg);
  }

  ImageGroup*   m_imageGroup;
  ImageGroupMsg m_msg;
};

void ImageGroup::notify(ImageGroupMsg message)
{
  std::for_each(m_impl->m_observers.begin(), m_impl->m_observers.end(), 
               Update(this, message));
}

bool ImageGroup::inqTarnished()
{
  CheckImageTarnish check = std::for_each(m_impl->m_imageList.begin(),
                                          m_impl->m_imageList.end(),
                                          CheckImageTarnish());
  return check.result();
}

